#line 1 "../main.ino"
#include <Arduino.h>
#line 1
void setup ();
void loop ();
void state_logic (void );
void output_logic (void );
void interrupt_service_routine_INT1 (void );
void write_outputs (void );
void update_time (void );
void datapath (void );
#line 1
/*	===========================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu 

	P12: Designing dedicated processors using microcontrollers (FSM + datapath)
	
Controlling a relay using an ultrasonic sensor  
 Documentation: https://digsys.upc.edu/csd/P12/Tap/Tap_control.html
 Target: Arduino Uno, chip ATmega328P
===========================================================================
This is an auto matic system for switching on and off a load when an obstacle
is detected in front of an ultrasonic sensor. 
The load may be any actuator: lamps, electro-valves, relays, etc.  

Planning: 
Hardware: Draw an schematic using Arduino and selecting pins as inputs 
and outputs.

Software
First ideas on using the typical sensor HC-SR04. Find a library for using 
the sensor or use functions already installed 
such "pulseIn()"
https://digsys.upc.edu/csd/P12/MAGNET/Aixeta/Control_Aixeta.html

Unfortunately "pulseIn()" function blocks the code execution when performing
the pulse duration measurement, so, it cannot be used in a FSM context that 
is running continuously as fast as possible without delays or loops. 

Thus, it is better to learn how to replicate pulseIn() function for measuring
pulse durations using interrupts:
https://roboticsbackend.com/arduino-pulsein-with-interrupts/
https://roboticsbackend.com/arduino-millis-vs-micros/

- Organise a FSM state diagram such:

- Draw the hardware and software diagram, naming the main blocks

- Name all the RAM variables required

- Use a similar P12 project to copy and adapt, for instance the temperature 
  meter based in Arduino https://digsys.upc.edu/csd/P12/Temp/Temp.html
=============================================================================*/

/*	===========================================================================
	Constants
	=========================================================================*/
const byte Echo_pin = 3; 	// This can be INT0 (PD2) or INT 1 (PD3)
const byte Trigger_pin = 4; // this can be any port pin

const byte Pilot_LED_pin = 7; 
const byte Tap_pin = 8;

// Constants will allow us to adjust how the system works simply changing 
// their values and recompiling. 

// CLK time is used for LED intermittency 
const unsigned long INTERMITTENCE_PERIOD = 500;  // milliseconds 

// Obstacle detection threshold
const unsigned long DISTANCE_THRESHOLD = 40; // 40 cm

const unsigned long TAP_TIMING_PERIOD = 20*INTERMITTENCE_PERIOD; //10 s 

// Distance measurement period: every 2 s
const unsigned long MEASUREMENT_PERIOD = 4*INTERMITTENCE_PERIOD; 

// FSM list of states
const char OBSTACLE_DETECT	= 'A'; 
const char START_TAP_TIMER	= 'B';
const char TAP_WATER_ON		= 'C';
const char STOP_TAP_TIMER	= 'D';


/*	===========================================================================
	Variables
	======================================================================== */

char	var_PS_flag; // '1' when detecting the start of the Echo pulse
char	var_PE_flag; // '1' when detecting the end of the Echo pulse

char	var_Obstacle_flag; // '1' when the obstacle is detected 

char	var_TT_flag; // '1' when the TAP_TIMING_PERIOD has elapsed
char	var_toggle_LED_flag; // '1' every INTERMITTENCE_PERIOD
char 	var_Meas_flag; // '1' every MEASUREMENT_PERIOD

char	var_current_state ;    // The internal memory of the FSM 

// outputs
char	var_Pilot_LED;
char	var_Vc;

//sensor interface
char	var_Trigger;
char	var_Echo; 

// Time calculations
unsigned long	var_previous_time_LED;
unsigned long	var_previous_time_Water;
unsigned long	var_previous_time_Meas;

// Distance variables
unsigned long	var_pulse_start_time;
unsigned long	var_pulse_end_time;
unsigned long	var_Pulse_Duration;  // Echo signal pulse duration 
unsigned long	var_Distance;  // Distance from sensor to obstacle (hand) 

/*	===========================================================================
	Initialise system
	=========================================================================*/
void setup () {
// Serial port 
    Serial.begin(9600);
	Serial.println("EETAC - CSD - P12 - Project Tap control");

// Configure inputs and outputs 
	pinMode(Echo_pin, INPUT);  
	pinMode(Trigger_pin, OUTPUT); 
	
	pinMode(Pilot_LED_pin, OUTPUT);  
	pinMode(Tap_pin, OUTPUT);	

//  Reset the sensor trigger signal 
    digitalWrite(Trigger_pin, LOW);   

//  Switch off the tap and the pilot LED
    digitalWrite(Tap_pin, LOW);   // Tap = 0;
    digitalWrite(Pilot_LED_pin, LOW);   // Pilot = 0;

// =========================================================
// Install the INT1 interrupt on event (change logic level)
	attachInterrupt(digitalPinToInterrupt(Echo_pin), interrupt_service_routine_INT1, CHANGE);
// =========================================================

//Initialise the FSM 
var_current_state = OBSTACLE_DETECT; 

// Initialise all the RAM variables
var_PS_flag 		= 0; 
var_PE_flag 		= 0; 
var_Obstacle_flag 	= 0;
var_TT_flag 		= 0; 
var_toggle_LED_flag	= 0; 
var_Meas_flag 		= 0;
var_Distance 		= 200;

// For calculating time
var_previous_time_LED 	= 0;
var_previous_time_Water	= 0;
var_previous_time_Meas 	= 0;

// Reset output values
var_Trigger = 0;
var_Pilot_LED = 0; 
var_Vc = 0;
var_Echo = 0;

}

/*	===========================================================================
	Infinite loop (FSM execution)
	=========================================================================*/
void loop() {       
	state_logic();    
	output_logic();
}


/*	===========================================================================
	State logic to determine where to go next
	======================================================================== */
void state_logic(void){

	switch (var_current_state) {
	case OBSTACLE_DETECT:	
		if (var_Obstacle_flag == 1) {	  // start the process
			  var_current_state = START_TAP_TIMER;
			  var_Obstacle_flag = 0;	  //Clear flag when used
		}
		break;

	case START_TAP_TIMER:	
		var_current_state = TAP_WATER_ON;       
		break;

	case TAP_WATER_ON:	// run for TAP_TIMING_PERIOD duration
		if (var_TT_flag == 1) {
			  var_current_state = STOP_TAP_TIMER;
			  var_TT_flag = 0;	  //Clear flag when used
		}
		break;

	case STOP_TAP_TIMER:	// go to measuring distances
		var_current_state = OBSTACLE_DETECT;     
		break;

	}	
}

/*	===========================================================================
	Output logic 
	======================================================================== */
void output_logic(void){

	update_time();  	
	
	switch (var_current_state) {
	case OBSTACLE_DETECT:
		var_Vc = 0; // electro-valve OFF	
		var_Pilot_LED = 0; 

		datapath();

		break;

	case START_TAP_TIMER:
			// initialise  
		var_TT_flag 		= 0; 
		var_toggle_LED_flag	= 0; 
			// Initialise the Water ON timings 
		var_previous_time_LED = millis();
		var_previous_time_Water = millis();

		break;

	case TAP_WATER_ON:
	
		if (var_toggle_LED_flag == 1){
			var_Pilot_LED = var_Pilot_LED^0b00000001; // bitwise NOT operation
			var_toggle_LED_flag = 0;
		}
		var_Vc = 1; // Drive the electro-valve 
		
		break;

	case STOP_TAP_TIMER: // Initialise distance measurements
		var_Distance = 200;	
		var_Vc = 0; // electro-valve OFF	
		var_Pilot_LED = 0; // LED OFF
			// Initialise the measurement timing  
		var_previous_time_Meas = millis();	

		break;	
	}
	write_outputs();
	
}

/*	===========================================================================
	ISR external interrupt INT1 pin 
This function is executed when a change (event ) is detected at Pin PD3 (INT1)
And this happens when the sensor switches HIGH and later LOW the Echo signal
	======================================================================== */	
void interrupt_service_routine_INT1 (void){
	var_Echo = digitalRead(Echo_pin); // read the input pin
	if (var_Echo == 1) 
		var_PS_flag = 1; // The INT1 has happened at the beginning of the pulse
	else 
		var_PE_flag = 1; // The interrupt has happened at the end of the pulse

}


/*	===========================================================================
	write_outputs
	======================================================================== */
void write_outputs(void){
   if (var_Vc == 1)
       digitalWrite(Tap_pin, HIGH); 
   else
       digitalWrite(Tap_pin, LOW);  
   
   if (var_Pilot_LED == 1)
       digitalWrite(Pilot_LED_pin, HIGH);  
   else
       digitalWrite(Pilot_LED_pin, LOW);    
}


/*	===========================================================================
	Update time to calculate the period actions : pilot LED toggle, 
	water ON duration, measurement period  
	======================================================================== */
void update_time(void){

static unsigned long var_current_time;    // current time in ms
static unsigned long var_time_elapsed_LED;   //  time elapsed since last flag
static unsigned long var_time_elapsed_Water;  
static unsigned long var_time_elapsed_Meas;  

  // From internal CLK: 
    var_current_time = millis();
    var_time_elapsed_LED =  var_current_time - var_previous_time_LED;  
    var_time_elapsed_Water =  var_current_time - var_previous_time_Water;  
    var_time_elapsed_Meas =  var_current_time - var_previous_time_Meas;  
	if(var_time_elapsed_LED > INTERMITTENCE_PERIOD) {
		var_toggle_LED_flag = 1;   
		var_previous_time_LED = var_current_time; 
	}
	
	if(var_time_elapsed_Water > TAP_TIMING_PERIOD) {
		var_TT_flag = 1;   
		var_previous_time_Water = var_current_time; 
	}
	if(var_time_elapsed_Meas > MEASUREMENT_PERIOD) {
		var_Meas_flag = 1;   
		var_previous_time_Meas = var_current_time; 
	}

		
}

/*	===========================================================================
	datapath function for measuring distance and detecting obstacles
	var_Obstacle_flag = 1 when the measured distance to the obstacle 
	is less than DISTANCE_THRESHOLD
		
	======================================================================== */
void datapath(void){

	if 	(var_Meas_flag == 1){
		// Trigger the ultrasonic sensor periodically		
		digitalWrite(Trigger_pin, HIGH);
		delayMicroseconds(5);
		digitalWrite(Trigger_pin, LOW);
		delayMicroseconds(3);		
		var_Meas_flag = 0;
	}

	//Distance measurement 
	if (var_PS_flag == 1){
		var_pulse_start_time = micros();
		var_PS_flag = 0;
	}
	if (var_PE_flag == 1){  //Calculate pulse duration in microseconds 
		var_pulse_end_time = micros();

		var_Pulse_Duration = var_pulse_end_time - var_pulse_start_time; 

// And calculate the obstacle distance in cm			
// The obstacle distance is calculated from the pulse duration measurement
// Sound speed Vs = 343 m/s; 
// Obstacle distance var_D(m) = (1/2)Tp(s)Vs(m/s)
// change units to cm and microseconds: 
// var_Distance is in cm; var_Pulse_Duration is in microseconds
// The constant in the equation is (1/58.1) cm/microsecond
		var_Distance = var_Pulse_Duration/58.1;			

// Transmit the measured Distance to the serial monitor 
		Serial.print("Distance: "); Serial.print(var_Distance); 
		Serial.println(" cm");
		
		var_PE_flag = 0;
	}
// Obstacle detection 
	if (var_Distance < DISTANCE_THRESHOLD){ 
		var_Obstacle_flag = 1;	// Change state to activate water flow									
	}		
 			 		
}
